# coding: utf-8
#
# Copyright 2012 Alexandr Emelin
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import cyclone.web
from cyclone.sse import SSEHandler

from twisted.internet import defer
from twisted.internet import base
from twisted.internet import reactor
from twisted.internet import task
from twisted.python import log


class ExtendedSSEHandler(SSEHandler):

    def reset_ping(self):
        self.del_ping()
        self.set_ping()

    def set_ping(self):
        """
        sets delayed call for this client
        """
        ping = defer.Deferred()
        ping.addCallback(self.send_ping)
        ping.addErrback(lambda err: log.err(err))
        self.ping = reactor.callLater(30, ping.callback, True)

    def del_ping(self):
        """
        cancels delayed call for this client
        """
        if hasattr(self, 'ping') and isinstance(self.ping, base.DelayedCall):
            if not self.ping.called and not self.ping.cancelled:
                self.ping.cancel()

    def send_ping(self, _ignored):
        """
        sends comment line to keep connection with client opened as mentioned here:
        https://developer.mozilla.org/en-US/docs/Server-sent_events/Using_server-sent_events
        """
        self.transport.write(": %s\n\n" % 'sse ping')
        if self.is_xhr():
            # we should close connection with XMLHttpRequest clients to make them reconnect
            self.unbind()
        else:
            self.set_ping()

    def is_xhr(self):
        return self.request.headers.get('X-Requested-With', None) == 'XMLHttpRequest'


class BroadcastHandler(ExtendedSSEHandler):

    sse_headers = {
        'Content-Type': 'text/event-stream; charset=utf-8',
        'Cache-Control': 'no-cache',
        'Connection': 'keep-alive',
        'Access-Control-Allow-Origin': '*'
    }

    def initialize(self):
        for name, value in self.sse_headers.iteritems():
            self.set_header(name, value)

    def get_channels(self):
        """
        retrieves channels client wants to subscribe
        """
        channels = self.get_arguments('channels')
        return channels

    def authorize(self):
        pass

    def bind(self):
        """
        called when new connection established 
        """
        log.msg(self.request.headers)
        self.authorize()
        self.write('\n\n')
        self.flush()
        self.application.broker.add_client(self)

    def unbind(self):
        """
        called when connection with client lost
        """
        self.application.broker.remove_client(self)


class StatsHandler(cyclone.web.RequestHandler):
    """
    current connection statistics
    """
    def get(self):
        self.write(str([(k, len(v)) for k, v in self.application.broker._channels.iteritems()]))
        self.finish()


class PublishHandler(cyclone.web.RequestHandler):
    """
    handler that listens to incoming messages and publishes new into HTTP broker
    """
    def post(self):
        key = self.get_argument("key", None)
        secret_key = self.application.broker.secret_key
        if secret_key and key != secret_key:
            raise cyclone.web.HTTPError(401)
        message = self.get_argument('message', None)
        channel = self.get_argument("channel", None)
        if channel and message:
            self.application.broker.publish(channel, message)
            self.set_header("Content-Type", "application/json")
            self.write({'status':'ok'})
        else:
            return cyclone.web.HTTPError(400)


# coding: utf-8
#
# Copyright 2012 Alexandr Emelin
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import sys

from twisted.application import internet
from twisted.application import service
from twisted.plugin import IPlugin
from twisted.python import usage
from twisted.python import reflect
from zope.interface import implements

from cyclone_sse.server import App

try:
    from twisted.internet import ssl
except ImportError:
    ssl_support = False
else:
    ssl_support = True


class Options(usage.Options):
    optParameters = [
        ["port", "p", 8888, "tcp port to listen on", int],
        ["listen", "l", "127.0.0.1", "interface to listen on"],
        ["app", "r", None, "custom application to run"],

        ["use-ssl", None, 0, "use ssl", int],
        ["ssl-port", None, 8443, "port to listen on for ssl", int],
        ["ssl-listen", None, "127.0.0.1", "interface to listen on for ssl"],
        ["ssl-cert", None, "server.crt", "ssl certificate"],
        ["ssl-key", None, "server.key", "ssl server key"],
        ["ssl-app", None, None, "ssl application (same as --app)"],
        ["ssl-appopts", None, None, "arguments to the ssl application"],

        ["broker", None, "http", "publish provider"],

        ["http-secret", None, "", "secret key to publish via HTTP"],

        ["redis-host", None, "127.0.0.1", "redis host"],
        ["redis-port", None, 6379, "redis port", int],
        
        ["amqp-host", None, "127.0.0.1", "amqp broker's host"],
        ["amqp-port", None, 5672, "amqp broker's port", int],
        ["amqp-vhost", None, "/", "amqp broker's virtual host"],
        ["amqp-spec", None, "rabbit.xml", "path to AMQP specification XML file"],
        ["amqp-username", None, "guest", "username"],
        ["amqp-password", None, "guest", "password"],
        ["amqp-exchange-name", None, "sse-cyclone-server", "amqp exchange name"],
        ["amqp-exchange-type", None, "direct", "amqp exchange type"],
        ["amqp-channel", None, 1, "amqp channel number", int],
    ]


class ServiceMaker(object):
    implements(service.IServiceMaker, IPlugin)
    tapname = "cyclone-sse"
    description = "SSE (EventSource) broadcasting server"
    options = Options

    def makeService(self, options):
        srv = service.MultiService()
        s = None

        if options["app"]:
            app = reflect.namedAny(options["app"])
        else:
            app = App

        # http
        s = internet.TCPServer(options["port"], app(options),
                               interface=options["listen"])
        s.setServiceParent(srv)

        # https
        if options["use-ssl"]:
            if ssl_support:
                s = internet.SSLServer(options["ssl-port"], app(options),
                                       ssl.DefaultOpenSSLContextFactory(
                                       options["ssl-key"],
                                       options["ssl-cert"]),
                                       interface=options["ssl-listen"])
                s.setServiceParent(srv)
            else:
                print("SSL is not supported. Please install PyOpenSSL.")

        if s is None:
            print("Nothing to do. Try --help")
            sys.exit(1)

        return srv

serviceMaker = ServiceMaker()


# coding: utf-8
#
# Copyright 2012 Alexandr Emelin
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import cyclone.web

from cyclone_sse.handlers import BroadcastHandler
from cyclone_sse.handlers import PublishHandler
from cyclone_sse.handlers import StatsHandler

from cyclone_sse.brokers import HttpBroker
from cyclone_sse.brokers import RedisBroker
from cyclone_sse.brokers import AmqpBroker


class App(cyclone.web.Application):
    def __init__(self, settings):
        handlers = [
            (r"/", BroadcastHandler),
            (r"/stats", StatsHandler),
        ]

        if settings["broker"] == 'amqp':
            broker = AmqpBroker
        elif settings["broker"] == 'redis':
            broker = RedisBroker
        else:
            broker = HttpBroker
            handlers.append((r"/publish", PublishHandler))

        self.broker = broker(settings)
        cyclone.web.Application.__init__(self, handlers)




# coding: utf-8
#
# Copyright 2012 Alexandr Emelin
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from twisted.internet import reactor
from twisted.internet import defer
from twisted.internet import protocol
from twisted.python import log

from txamqp.protocol import AMQClient
from txamqp.client import TwistedDelegate
from txamqp.content import Content
import txamqp


class AmqpSubscriberProtocol(AMQClient):

    def connectionMade(self):
        AMQClient.connectionMade(self)
        df = self.connect()
        df.addCallback(self.connect_success)

    def connect_success(self, channel):
        self.connected = 1
        self.channelReady()
        self.factory.addConnection(self)
        
    def channelReady(self):
        pass

    def connection_error(self, err):
        log.err(err)
        self.factory.continueTrying = False
        self.transport.loseConnection()

    @defer.inlineCallbacks
    def connect(self):
        try:
            yield self.start({'LOGIN': self.factory.username,
                              'PASSWORD': self.factory.password})
        except Exception, e:
            self.connection_error(e)
            defer.returnValue(None)

        try:
            self.chan = yield self.channel(self.factory.chan)
        except Exception, e:
            self.connection_error(e)
            defer.returnValue(None)

        try:
            yield self.chan.channel_open()
        except Exception, e:
            self.connection_error(e)
            defer.returnValue(None)

        defer.returnValue(self.chan)

    def connectionLost(self, reason):
        AMQClient.connectionLost(self, reason)
        self.connected = 0
        self.factory.delConnection(self)

    @defer.inlineCallbacks
    def consume(self, routing_key):
        # declaring exchange
        try:
            yield self.chan.exchange_declare(exchange=self.factory.exchange_name,
                                             type=self.factory.exchange_type)
        except Exception, e:
            log.err(e)
            defer.returnValue(None)    

        # declaring queue
        try:
            result = yield self.chan.queue_declare(exclusive=True)
            queuename = result.fields[0]
        except Exception, e:
            log.err(e)
            defer.returnValue(None)

        # binding queue to exchange with routing_key = queue name
        try:
            yield self.chan.queue_bind(exchange=self.factory.exchange_name,
                                       queue=queuename,
                                       routing_key=routing_key)
        except Exception, e:
            log.err(e)
            defer.returnValue(None)    

        # subscribing on queue
        try:
            yield self.chan.basic_consume(queue=queuename,
                                          no_ack=True,
                                          consumer_tag=routing_key)
        except Exception, e:
            log.err(e)
            defer.returnValue(None)

        queue = yield self.queue(routing_key)

        while True:
            log.msg('consuming %s' % routing_key)
            try:
                msg = yield queue.get()
            except Exception, e:
                defer.returnValue(None)
            self.messageReceived(None, msg.routing_key, msg.content.body)

    @defer.inlineCallbacks
    def cancel(self, queue_name):
        try:
            yield self.chan.basic_cancel(consumer_tag=queue_name)
            print '%s stopped' % queue_name
        except Exception, e:
            log.err(e)
            defer.returnValue(None)

    def messageReceived(self, pattern, channel, message):
        pass


class AmqpSubscriberFactory(protocol.ReconnectingClientFactory):
    maxDelay = 120
    continueTrying = True
    protocol = AmqpSubscriberProtocol

    def __init__(self, spec_file=None, vhost=None,
                 host=None, port=None, username=None,
                 password=None, exchange_name="",
                 exchange_type="fanout", channel=None):
        spec_file = spec_file or 'extras/rabbitmq-specification.xml'
        self.spec = txamqp.spec.load(spec_file)
        self.username = username or 'guest'
        self.password = password or 'guest'
        self.vhost = vhost or '/'
        self.chan = channel or 1
        self.delegate = TwistedDelegate()
        self.exchange_name = exchange_name
        self.exchange_type = exchange_type

    def buildProtocol(self, addr):
        p = self.protocol(self.delegate, self.vhost, self.spec)
        p.factory = self # Tell the protocol about this factory.
        # Reset the reconnection delay since we're connected now.
        self.resetDelay()
        return p

    def addConnection(self, conn):
        pass

    def delConnection(self, conn):
        pass



import sys

requires = ["twisted", "cyclone", "txAMQP"]

try:
    # to avoid installation problems on centos 5
    # which has no distributions for OpenSSL 0.9.8f
    from distutils.version import StrictVersion
    import platform
    os, version, _type = platform.linux_distribution()
    if os == "CentOS" and StrictVersion(version) < StrictVersion('6.0'):
        requires.insert(0, "pyopenssl==0.12")
except:
    pass

from distutils import log
from distutils.core import setup

# PyPy and setuptools don't get along too well, yet.
if sys.subversion[0].lower().startswith('pypy'):
    from distutils.core import setup
    extra = dict(requires=requires)
else:
    from setuptools import setup
    extra = dict(install_requires=requires)


setup(
    name="cyclone-sse",
    version="0.7.0",
    author="Alexandr Emelin",
    author_email="frvzmb@gmail.com",
    url="https://github.com/FZambia/cyclone-sse",
    license="http://www.apache.org/licenses/LICENSE-2.0",
    description="EventSource (SSE, Server-Sent-Events) broadcasting server with channel support, Last-Event-Id etc, based on top of cyclone web server",
    keywords="python non-blocking twisted cyclone sse eventsource broadcast http redis amqp rabbitmq",
    packages=["cyclone_sse", "twisted.plugins"],
    #package_data={"twisted": ["plugins/cyclonesse_plugin.py"]},
    **extra
)

# Make Twisted regenerate the dropin.cache, if possible.  This is necessary
# because in a site-wide install, dropin.cache cannot be rewritten by
# normal users.
try:
    from twisted.plugin import IPlugin, getPlugins
except ImportError:
    pass
else:
    list(getPlugins(IPlugin))



# coding: utf-8
#
# Copyright 2012 Alexandr Emelin
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import uuid

from twisted.internet import defer
from twisted.internet import reactor
from twisted.internet import task
from twisted.python import log

import cyclone.redis
from cyclone_sse.amqp import AmqpSubscriberFactory
from cyclone_sse.amqp import AmqpSubscriberProtocol


class Broker(object):
    """
    base class, which goal is to keep state of connections and to
    broadcast new messages to all clients of certain channel.
    """
    def __init__(self, settings, name='unknown'):
        self._name = name
        self._source = None
        self._clients = {}
        self._channels = {}
        self._cache = []
        self._cache_size = 50
        self._settings = settings
        self.setup(self._settings)

    def setup(self, settings):
        self.connect(settings)

    def connect(self, settings):
        raise NotImplementedError('please, provide implementation for connect method')

    def _subscribe(self, channel):
        if channel not in self._channels:
            self._channels[channel] = {}
            log.msg("Subscribing entire server to %s" % channel)
            if self._source:
                self.subscribe(channel)

    def _unsubscribe(self, channel):
        if channel in self._channels:
            del self._channels[channel]
            log.msg('Unsubscribing entire server from channel %s' % channel)
            if self._source:
                self.unsubscribe(channel)

    def add_client(self, client):
        """
        registers new client in broker
        """
        client.uid = str(uuid.uuid4())
        client.channels = client.get_channels()
        if not client.channels:
            raise cyclone.web.HTTPError(400)

        self._clients[client.uid] = client
        client.set_ping()

        for channel in client.channels:
            self._subscribe(channel)
            self._channels[channel][client.uid] = 1
            log.msg("Client %s subscribed to %s" % (client.request.remote_ip, channel))

        self.send_cache(client)

    def remove_client(self, client):
        """
        unregisters client
        """
        if client.uid in self._clients:
            del self._clients[client.uid]
            client.del_ping()
            try:
                client.flush()
                client.finish()
            except defer.AlreadyCalledError:
                # client connection has already been finished
                pass

        # channels that have no clients
        empty = []

        # removing client from channels
        for channel in client.channels:
            clients = self._channels.get(channel, None)
            if clients and client.uid in clients:
                del clients[client.uid]
                log.msg('Unsubscribing client from channel %s' % channel)
            if not clients:
                empty.append(channel)

        # clean server subscription
        for channel in empty:
            self._unsubscribe(channel)

    def subscribe(self, channel):
        raise NotImplementedError('please, provide implementation for subscribe method')

    def unsubscribe(self, channel):
        raise NotImplementedError('please, provide implementation for unsubscribe method')

    def broadcast(self, pattern, channel, message):
        """
        sends message to all clients of certain channel
        """
        if self.is_pattern_blocked(pattern):
            return True
        clients = self._channels.get(channel, None)

        if clients:
            args = (str(len(clients)), pattern, channel, message)
            log.msg('BROADCASTING to %s clients: pattern: %s, channel: %s, message: %s' % args)            

            # put this message into cache
            eid = str(uuid.uuid4())
            self.update_cache(eid, channel, message)

            # sent message to all clients
            for uid in clients:
                client = self._clients.get(uid, None)
                if client:
                    self.send_event(client, message, eid=eid)
                    client.reset_ping()

    def is_pattern_blocked(self, pattern):
        """
        must return true if we do not want to send messages of this pattern
        """
        return False

    def send_event(self, client, message, eid):
        client.sendEvent(message, eid=eid)
        if client.is_xhr():
            client.unbind()

    def send_cache(self, client):
        """
        sends missed messages to client
        """
        last_event_id = client.request.headers.get('Last-Event-Id', None)
        if last_event_id:
            i = 0
            for i, msg in enumerate(self._cache, 1):
                if msg['eid'] == last_event_id:
                    break

            for item in self._cache[i:]:
                if item['channel'] in client.get_channels():
                    self.send_event(client, item['message'], item['eid'])
                    if client.is_xhr():
                        break

    def update_cache(self, eid, channel, message):
        """
        adds message into broker cache
        """
        self._cache.append({
            'eid': eid,
            'channel': channel,
            'message': message
        })
        if len(self._cache) > self._cache_size:
            self._cache = self._cache[-self._cache_size:]


class HttpBroker(Broker):

    def __init__(self, *args, **kwargs):
        super(HttpBroker, self).__init__(*args, **kwargs)

    def connect(self, settings):
        self.secret_key = settings["http-secret"]
        self.queue = defer.DeferredQueue()
        self._source = True
        self.start_consuming()

    @defer.inlineCallbacks
    def start_consuming(self):
        while True:
            msg = yield self.queue.get()
            self.broadcast(None, msg['channel'], msg['message'])

    def subscribe(self, channel):
        pass

    def unsubscribe(self, channel):
        pass

    def publish(self, channel, message):
        if channel in self._channels:
            self.queue.put({'channel': channel, 'message': message})


class RedisBroadcastProtocol(cyclone.redis.SubscriberProtocol):
    def messageReceived(self, pattern, channel, message):
        # When new messages are published to Redis channels or patterns,
        # they are broadcasted to all HTTP clients subscribed to those
        # channels.
        self.factory.broker.broadcast(pattern, channel, message)

    def connectionMade(self):
        self.factory.broker._source = self
        # If we lost connection with Redis during operation, we
        # re-subscribe to all channels once the connection is re-established.
        for channel in self.factory.broker._channels:
            self.factory.broker._subscribe(channel)

    def connectionLost(self, why):
        self.factory.broker._source = None


class RedisBroker(Broker):
    """
    listens Redis channels and broadcasts to clients
    """
    def connect(self, settings):
        # PubSub client connection
        qf = cyclone.redis.SubscriberFactory()
        qf.broker = self
        qf.maxDelay = 20
        qf.protocol = RedisBroadcastProtocol
        reactor.connectTCP(settings["redis-host"], settings["redis-port"], qf) 

    def is_pattern_blocked(self, pattern):
        return pattern in ['unsubscribe', 'subscribe']

    def subscribe(self, channel):
        if "*" in channel:
            self._source.psubscribe(channel)
        else:
            self._source.subscribe(channel)

    def unsubscribe(self, channel):
        if "*" in channel:
            self._source.punsubscribe(channel)
        else:
            self._source.unsubscribe(channel)


class AmqpBroadcastProtocol(AmqpSubscriberProtocol):
    def messageReceived(self, pattern, channel, message):
        # broadcast this message to all HTTP clients
        self.factory.broker.broadcast(pattern, channel, message)

    def channelReady(self):
        self.factory.broker._source = self
        # If we lost connection during operation, we
        # re-subscribe to all channels once the connection is re-established.
        for channel in self.factory.broker._channels:
            self.factory.broker._subscribe(channel)   

    def connectionLost(self, why):
        self.factory.broker._source = None
        AmqpSubscriberProtocol.connectionLost(self, why)


class AmqpBroker(Broker):
    def connect(self, settings):
        # PubSub client connection
        qf = AmqpSubscriberFactory(spec_file=settings["amqp-spec"],
                                   vhost=settings["amqp-vhost"],
                                   username=settings["amqp-username"],
                                   password=settings["amqp-password"],
                                   exchange_name=settings["amqp-exchange-name"],
                                   exchange_type=settings["amqp-exchange-type"],
                                   channel=settings["amqp-channel"])
        qf.broker = self
        qf.protocol = AmqpBroadcastProtocol
        reactor.connectTCP(settings["amqp-host"], settings["amqp-port"], qf)

    def subscribe(self, channel):
        self._source.consume(channel)

    def unsubscribe(self, channel):
        self._source.cancel(channel)
